@@ -1,11 +1,11 @@
+from json import load
 import snowflake.connector
 import pickle
 import os, sys
-from databricks import sql
+# from databricks import sql
 # from snowflake.sqlalchemy import URL
 from sqlalchemy import create_engine
 # import pandas.io.sql as pdsql
-import logging
 from datetime import datetime as dt 
 import os.path
 from os import path
@@ -15,10 +15,12 @@ from snowflake.connector.pandas_tools import pd_writer
 import pytz
 import hashlib
 import getpass
-import base64, boto3, json, redshift_connector
-from cryptography.hazmat.backends import default_backend
-from cryptography.hazmat.primitives import serialization
-from functional_functions.ff_classes import FBI_S3, DBX_sql
+# import json
+import redshift_connector
+from functional_functions.ff_classes import FBI_S3, DBX_sql, AWS_Secrets
+
+import logging
+# logging.basicConfig(level=logging.DEBUG, format='%(asctime)s :: %(name)s -- %(funcName)s() :: %(levelname)s :: %(message)s')
 
 # import jaydebeapi as jay
 
@@ -47,11 +49,10 @@ def help():
     save_pickle()
     batch_start()
     batch_update()
-    aws_secrets_manager_getvalues()
-    decrypt_aws_private_key()
-    decode_snowflake_username()
-    _decode_string()
-    read_value()
+    grant_all_permissions_dbx()
+    load_via_spark_dbx()
+    ( ONLY for DBX ) query_method_by_env_dbx()
+    ( more func available in ff_classes )
     '''
     
     print('''
@@ -71,14 +72,13 @@ def help():
     save_pickle()
     batch_start()
     batch_update()
-    aws_secrets_manager_getvalues()
-    decrypt_aws_private_key()
-    decode_snowflake_username()
-    _decode_string()
-    read_value()
+    grant_all_permissions_dbx()
+    load_via_spark_dbx()
+    ( ONLY for DBX ) query_method_by_env_dbx()
+    ( more func available in ff_classes )
     ''')
 
-def load_via_sql_snowflake(load_df, tbl_name, secret_name = 'fbi_snowflake_creds', if_exists='replace', creds=None, test_mode=None):
+def load_via_sql_snowflake(load_df, tbl_name, if_exists='replace', creds=None, test_mode=None):
     '''
     This is the function that load pd df direct via SQL instead of a CSV fashion.
     Currently designed to be used with sandbox or current designed database via settings file.
@@ -101,59 +101,10 @@ def load_via_sql_snowflake(load_df, tbl_name, secret_name = 'fbi_snowflake_creds
         tbl_name = tbl_name.upper()
 
 
-    print('loading tbl ' + tbl_name)
+    print('loading tbl ' + tbl_name + ' in snowflake')
 
-    secrets = aws_secrets_manager_getvalues(secret_name, key_id = None, access_key = None)
-    KEY_PREFIX = '-----BEGIN ENCRYPTED PRIVATE KEY-----\n'
-    KEY_POSTFIX = '\n-----END ENCRYPTED PRIVATE KEY-----\n'
-    pkey = KEY_PREFIX + '\n'.join(secrets['snowflake_secret_key'].split(' ')) + KEY_POSTFIX
-    snowflakePassPhrase = secrets['snowflake_pass_phrase']
-
-    password = _decode_string(snowflakePassPhrase).replace("\n", "")
-    password_to_byte = bytes(password, 'utf8')
-
-    private_key_to_byte = bytes(pkey, 'utf8')
-
-    p_key = serialization.load_pem_private_key(
-                    private_key_to_byte,
-                    password=password_to_byte,
-                    backend=default_backend()
-            )
-
-    pkb = p_key.private_bytes(
-                encoding=serialization.Encoding.DER,
-                format=serialization.PrivateFormat.PKCS8,
-                encryption_algorithm=serialization.NoEncryption()
-            )
-
-    snowflakeUsername = secrets['snowflake_usn']
-    username = _decode_string(snowflakeUsername).replace("\n", "")
-
-    if creds is None:
-        creds = {
-            'usr' : username,
-            'pkb' : pkb,
-            'role' : os.environ.get('SNOWLAKE_ROLE'),
-            'warehouse_name' : os.environ.get('SNOWFLAKE_WAREHOUSE_NAME'),
-            'db_name' : os.environ.get('SNOWFLAKE_DB_NAME'),
-            'schema' : os.environ.get('SNOWFLAKE_SCHEMA'),
-        }
-
-        #Use settings if ENV vars are not set up
-        if all(value is None for value in [creds['role'], creds['warehouse_name'],creds['db_name'], creds['schema']]):
-            creds = {
-                'usr' : username,
-                'pkb' : pkb,
-                'role' : settings.SNOWFLAKE_FPA['role'],
-                'warehouse_name' : settings.SNOWFLAKE_FPA['warehouse_name'],
-                'db_name' : settings.SNOWFLAKE_FPA['db_name'],
-                'schema' : settings.SNOWFLAKE_FPA['schema']
-            }
-    
-    try: 
-        schema = creds['schema']
-    except KeyError:
-        schema = 'FPA_SANDBOX'
+    creds_dict = creds if creds else AWS_Secrets().get_snowflake_secrets()
+    schema = creds_dict['schema']
 
     #Usually this only happens when reading in from a CSV, but better to be safe than sorry
     load_df.replace(abs(np.inf),0,inplace=True)
@@ -179,61 +130,93 @@ def load_via_sql_snowflake(load_df, tbl_name, secret_name = 'fbi_snowflake_creds
 
     load_df.columns = map(fix_column_name, load_df.columns)
 
-    conn = get_snowflake_connection(**creds) #, engine
+    conn = get_snowflake_connection(**creds_dict) #, engine
     engine = create_engine(f"snowflake://gl11689.us-east-1.snowflakecomputing.com", creator=lambda: conn)
 
     if if_exists == 'replace':
         print('dropping existing table')
         engine.connect().execute('drop table if exists ' + schema + '.' + tbl_name)
 
-    print('loading...')
+    # print('loading...')
     # load_df.to_sql(tbl_name, con=conn, index=False, if_exists='append', method=pd_writer)
     load_df.to_sql(tbl_name, con=engine.connect(), if_exists='append', index=False, method=pd_writer)
     print(f'{tbl_name} has been updated in SNOWFLAKES')
 
     # conn.close()
 
-def get_logger(filename):
+def get_logger(name, dirpath=None, level=logging.INFO):
     '''
-    The purpose of this function is to create a somewhat standard logger for when you need to add logging to your script.
-    The function will check your provided filepath for an existing logs folder, and will create one if it does not exist.
-    
-    sample syntax on your files:
-    
-    dir_path = os.path.dirname(os.path.realpath(__file__))
-
-    logging = get_logger(dir_path)
-    logger = logging.getLogger(__name__)
-
+        updated version for get_logger()
+        args:
+            -- name: suppose to be the name of the .py file. it only control the %(name) in logging formatter
+            -- dirpath: the path to log file; will create /log dir at the same directory
+                        in order to create logs at the same level of main.py file, having loggerconf.py as a wrapper;
+                        loggerconf.py file is at the same directory with main.py file, when import find_logger() from loggerconf; the directory remians the same
     '''
-    if sys.platform == "win32":
-        full_path = filename + r'\logs'
-        if path.exists(full_path):
-            print('found logs file, will be creating new logger in there')
-        else:
-            os.mkdir(filename + r'\logs')
-            print('You don\'t seem to have a logs file, what a shame. I\'ll go ahead and make one for you')
-    
-        logging.basicConfig(filename=full_path + r'\logger_' + dt.now().strftime('%Y-%m-%d %H.%M.%S') + '.log',
-            format='%(asctime)s :: %(name)s :: %(levelname)s :: %(message)s', 
-            filemode='a', 
-            datefmt='%H:%M:%S',
-            level=logging.INFO)
+
+    if dirpath == None:
+        dirpath = os.path.dirname(os.path.realpath(__file__))
+
+    if os.environ.get('environment') == 'databricks':
+        try:
+            logger = logging.getLogger(name)
+            # formatter = logging.Formatter("%(asctime)s %(levelname)s \t[%(filename)s:%(lineno)s - %(funcName)s()] %(message)s")	
+            formatter = logging.Formatter('%(asctime)s :: %(name)s -- %(funcName)s() :: %(levelname)s :: %(message)s')
+            logger.setLevel(level)
+
+            #add normal steam handler to display logs on screen
+            io_log_handler = logging.StreamHandler()
+            logger.addHandler(io_log_handler)
+
+            for handler in logger.handlers:
+                handler.setFormatter(formatter)
+        except Exception as e:
+            print(f"In Databricks Environment, cannot load logger, error: {str(e)}.")
+        return logger
+
     else:
-        full_path = filename + '/logs'
-        if path.exists(full_path):
-            print('found logs file, will be creating new logger in there')
+        if sys.platform == "win32":
+            full_path = dirpath + r'\logs'
+            try:
+                if path.exists(full_path):
+                    # print('found logs file, will be creating new logger in there')
+                    pass
+                else:
+                    os.mkdir(dirpath + r'\logs')
+                    # print('You don\'t seem to have a logs file, what a shame. I\'ll go ahead and make one for you')
+        
+                logging.basicConfig(filename=full_path + r'\logger_' + dt.now().strftime('%Y-%m-%d %H.%M.%S') + '.log',
+                    format='%(asctime)s :: %(name)s :: %(levelname)s :: %(message)s', 
+                    filemode='a', 
+                    datefmt='%H:%M:%S',
+                    level=level)
+            except Exception as e:
+                print('could not find or create logs dir, please create manually or double check permission')
+                print(str(e))
         else:
-            os.mkdir(filename + '/logs')
-            print('You don\'t seem to have a logs file, what a shame. I\'ll go ahead and make one for you')
-        logging.basicConfig(filename=full_path + '/logger_' + str(dt.now()) + '.log',
-            format='%(asctime)s :: %(name)s :: %(levelname)s :: %(message)s',
-            filemode='a',
-            datefmt='%H:%M:%S',
-            level=logging.INFO)
-    return logging
+            full_path = dirpath + '/logs'
+            try:
+                if path.exists(full_path):
+                    # print('found logs file, will be creating new logger in there')
+                    pass
+                else:
+                    os.mkdir(dirpath + '/logs')
+                    # print('You don\'t seem to have a logs file, what a shame. I\'ll go ahead and make one for you')
+                logging.basicConfig(filename=full_path + '/logger_' + str(dt.now()) + '.log',
+                    format='%(asctime)s :: %(name)s -- %(funcName)s() :: %(levelname)s :: %(message)s',
+                    filemode='a',
+                    datefmt='%H:%M:%S',
+                    level=level)
+            except Exception as e:
+                print('could not find or create logs dir, please create manually or double check permission')
+                print(str(e))
 
-def get_snowflake_connection(usr, pkb, role, warehouse_name, db_name=None, schema=None):
+        logger = logging.getLogger(name)
+        # io_log_handler = logging.StreamHandler()
+        # logger.addHandler(io_log_handler)
+        return logger
+
+def get_snowflake_connection(usr, pkb, role, warehouse_name, db=None, schema=None):
     '''
     The purpose of this function is to get a snowflake connection using credentials, usually stored in
     a settings file or env vars.
@@ -254,69 +237,23 @@ def get_snowflake_connection(usr, pkb, role, warehouse_name, db_name=None, schem
 
     '''
 
-    if db_name is None:
-        db_name = 'PC_STITCH_DB'
-    
-    if schema is None:
-        conn = snowflake.connector.connect(
-                            user=usr,
-                            account="gl11689.us-east-1",
-                            private_key=pkb,
-                            role = role,
-                            warehouse=warehouse_name,
-                            database=db_name,
-                            schema = 'NETSUITE_REPORTING'
-                            )
-    else:
-            conn = snowflake.connector.connect(
-                            user=usr,
-                            account="gl11689.us-east-1",
-                            private_key=pkb,
-                            role = role,
-                            warehouse=warehouse_name,
-                            database=db_name,
-                            schema = schema
-                            )
-    
+    db_name = db if db else 'PC_STITCH_DB'
+    schema_name = schema if schema else 'NETSUITE_REPORTING'
 
+    conn = snowflake.connector.connect(
+                    user=usr,
+                    account="gl11689.us-east-1",
+                    private_key=pkb,
+                    role = role,
+                    warehouse=warehouse_name,
+                    database=db_name,
+                    schema=schema_name
+                    )
     #conn.cursor().execute("USE role {}".format(role))
     
     return conn
 
-def decrypt_aws_private_key(secret_name):
-    secrets = aws_secrets_manager_getvalues(secret_name, key_id = None, access_key = None)
-    KEY_PREFIX = '-----BEGIN ENCRYPTED PRIVATE KEY-----\n'
-    KEY_POSTFIX = '\n-----END ENCRYPTED PRIVATE KEY-----\n'
-    pkey = KEY_PREFIX + '\n'.join(secrets['snowflake_secret_key'].split(' ')) + KEY_POSTFIX
-    snowflakePassPhrase = secrets['snowflake_pass_phrase']
-
-    password = _decode_string(snowflakePassPhrase).replace("\n", "")
-    password_to_byte = bytes(password, 'utf8')
-
-    private_key_to_byte = bytes(pkey, 'utf8')
-
-    p_key = serialization.load_pem_private_key(
-                    private_key_to_byte,
-                    password=password_to_byte,
-                    backend=default_backend()
-            )
-
-    pkb = p_key.private_bytes(
-                encoding=serialization.Encoding.DER,
-                format=serialization.PrivateFormat.PKCS8,
-                encryption_algorithm=serialization.NoEncryption()
-            )
-    
-    return pkb, secrets
-
-def decode_snowflake_username(secrets):
-    
-    snowflakeUsername = secrets['snowflake_usn']
-    username = _decode_string(snowflakeUsername).replace("\n", "")
-
-    return username
-
-def query_snowflake(query, secret_name='fbi_snowflake_creds', creds_dict=None):
+def query_snowflake(query, creds=None):
     '''
     This funky func is meant to query snowflake and cut out the middle man. Ideally it follows a cred or settings file
     structure. Will open a connection, query snowflake, and close connection and return dataframe
@@ -342,37 +279,13 @@ def query_snowflake(query, secret_name='fbi_snowflake_creds', creds_dict=None):
 
     '''
     
-    pkb, secrets = decrypt_aws_private_key(secret_name)
-    username = decode_snowflake_username(secrets)
+    creds_dict = creds if creds else AWS_Secrets().get_snowflake_secrets()
 
-
-
-    if creds_dict is None:
-        creds_dict = {
-            'usr' : username,
-            'pkb' : pkb,
-            'role' : os.environ.get('SNOWLAKE_ROLE'),
-            'warehouse_name' : os.environ.get('SNOWFLAKE_WAREHOUSE_NAME'),
-            'db_name' : os.environ.get('SNOWFLAKE_DB_NAME'),
-            'schema' : os.environ.get('SNOWFLAKE_SCHEMA'),
-        }
-
-        #Use settings if ENV vars are not set up
-        if all(value is None for value in [creds_dict['role'], creds_dict['warehouse_name'],creds_dict['db_name'], creds_dict['schema']]):
-            creds_dict = {
-                'usr' : username,
-                'pkb' : pkb,
-                'role' : settings.SNOWFLAKE_FPA['role'],
-                'warehouse_name' : settings.SNOWFLAKE_FPA['warehouse_name'],
-                'db_name' : settings.SNOWFLAKE_FPA['db_name'],
-                'schema' : settings.SNOWFLAKE_FPA['schema']
-            }
-    # get_snowflake_connection(warehouse_name, db_name=None, schema=None, usr, pkb)
     conn = get_snowflake_connection(**creds_dict)
 
     resp =  conn.cursor().execute(query).fetch_pandas_all()
 
-    conn.close()
+    # conn.close()
 
     return resp
 
@@ -439,7 +352,7 @@ def query_redshift(query, dsn_dict=None):
     return resp
 
 
-def query_databricks(query):
+def query_databricks(query, full_table = False, use_service_account = False):
     """
     This function is intended to enable users to connect and query from Databricks tables!
 
@@ -448,23 +361,20 @@ def query_databricks(query):
     'http_path',
     'access_token'
 
-    each of these can be gathered from Databricks by going to:
-    (For server_hostname and http_path)
-    Databricks --> SQL --> SQL Endpoints --> Click on Endpoint --> Connection Details
-
-    (For access_token)
-    Databricks --> SQL --> SQL Endpoints --> Click on Endpoint --> Connection Details --> Create a personal access token
-
     parameters
     --
     query : str, the SQL query being used
-    databricks_dict : dict, the dictionary with connection creds, see settings.py.sample REDSHIFT_SVC_ACCT for details or see creds.env.sample for ENV var names
-
-    if no dict is provided, will default to calling settings.py creds
-
+    full_table : bool; default False, only query top 10k rows on local to save query time; set to True if need full table
+    use_service_account: bool; default False, use personal account on local; automatically use service account on dbx
     """
-
-    dbx_sql = DBX_sql()
+    if use_service_account == True or os.environ.get('environment') == 'databricks':
+        dbx_host, dbx_path, dbx_token = AWS_Secrets().get_dbx_secrets()
+        dbx_sql = DBX_sql(server_hostname=dbx_host, http_path=dbx_path, access_token=dbx_token)
+    else:
+        dbx_sql = DBX_sql()
+    if full_table == False and os.environ.get('environment') != 'databricks':
+        if 'limit' not in query.lower():
+            query = query + " limit 10000"
     return dbx_sql.query_table(query)
 
 def save_pickle(df, file_name, folder_name=None, file_path_option=None):
@@ -562,7 +472,7 @@ def load_pickle(file_name, load_date=None, folder_name=None, file_path_option=No
     
     return df
 
-def batch_start(test_mode, script_name, secret_name='fbi_snowflake_creds', creds=None):
+def batch_start(test_mode, script_name, creds=None):
     '''
     This function is built to natively load into the FBI's batch table. The batch table is meant to
     track progress of script runs, this functionality will exist until a better solution is implemented.
@@ -583,39 +493,19 @@ def batch_start(test_mode, script_name, secret_name='fbi_snowflake_creds', creds
     dt_now = pytz.timezone("US/Eastern").localize(dt.now())
     hash_id = hashlib.md5(str(dt.now()).encode('utf-8')).hexdigest()
 
-    pkb, secrets = decrypt_aws_private_key(secret_name)
-    username = decode_snowflake_username(secrets)
-
-    usn_pkb = {
-                'usr': username,
-                'pkb': pkb
-            }
-
-    if creds is None:
-        creds = {
-            'role' : os.environ.get('SNOWLAKE_ROLE'),
-            'warehouse_name' : os.environ.get('SNOWFLAKE_WAREHOUSE_NAME'),
-            'db_name' : os.environ.get('SNOWFLAKE_DB_NAME'),
-            'schema' : os.environ.get('SNOWFLAKE_SCHEMA')
-        }
-
-        if all(value is None for value in creds.values()):
-            creds=settings.SNOWFLAKE_FPA
-    
-    usn_pkb.update(creds)
-    creds = usn_pkb
+    creds_dict = creds if creds else AWS_Secrets().get_snowflake_secrets()
             
-    conn = get_snowflake_connection(**creds)
+    conn = get_snowflake_connection(**creds_dict)
 
     q = """INSERT INTO fpa_sandbox.batch_table (batch_status,start_time,batch_hash,test_mode,script_name, user_running) 
                 VALUES ('Running...','{}','{}','{}','{}','{}');""".format(dt_now,hash_id,test_mode,script_name,user_running)
 
     conn.cursor().execute(q)
-    conn.close()
+    # conn.close()
 
     return hash_id
 
-def batch_update(status, hash_id, secret_name='fbi_snowflake_creds', creds=None):
+def batch_update(status, hash_id, creds=None):
     '''
     This function is built to update batch_table status to either finished or failed. 
     batch_start() must be run before this function can be called.
@@ -627,88 +517,89 @@ def batch_update(status, hash_id, secret_name='fbi_snowflake_creds', creds=None)
     hash_id - hash_id that was returned from batch_start()
     '''
 
-    pkb, secrets = decrypt_aws_private_key(secret_name)
-    username = decode_snowflake_username(secrets)
+    creds_dict = creds if creds else AWS_Secrets().get_snowflake_secrets()
 
-    usn_pkb = {
-                'usr': username,
-                'pkb': pkb
-            }
-
-    if creds is None:
-        creds = {
-            'role' : os.environ.get('SNOWLAKE_ROLE'),
-            'warehouse_name' : os.environ.get('SNOWFLAKE_WAREHOUSE_NAME'),
-            'db_name' : os.environ.get('SNOWFLAKE_DB_NAME'),
-            'schema' : os.environ.get('SNOWFLAKE_SCHEMA')
-        }
-
-        if all(value is None for value in creds.values()):
-            creds=settings.SNOWFLAKE_FPA
-    
-    usn_pkb.update(creds)
-    creds = usn_pkb
-
-    conn = get_snowflake_connection(**creds)
+    conn = get_snowflake_connection(**creds_dict)
     
     q = "UPDATE fpa_sandbox.batch_table SET batch_status = '{}', end_time = CURRENT_TIMESTAMP() WHERE batch_hash = '{}' ".format(status,hash_id)
 
     conn.cursor().execute(q)
-    conn.close()
+    # conn.close()
 
-def aws_secrets_manager_getvalues(secret_name, key_id = None, access_key = None):
-    '''
-    This function's purpose is to grab the aws secrets from the secrets manager
-    '''
-    try:
-        key_id = settings.AWS_SECRETS_MANAGER_CREDS['AWS_ACCESS_KEY_ID']
-        access_key = settings.AWS_SECRETS_MANAGER_CREDS['AWS_SECRET_ACCESS_KEY']
-    except: 
-        if key_id is None or access_key is None:
-            key_id = os.getenv("AWS_ACCESS_KEY_ID")
-            access_key = os.getenv("AWS_SECRET_ACCESS_KEY")
+def grant_table_permissions_dbx(tbl_name=None):
+    """
+        grant all permissions for FBI Team on all tables in finance_accounting catalog
+        if tbl_name given, only grant permission to that table
+        NOTE: all table being created through dbx_sql.create_or_replace_table() / load_via_sql_dbx() will grant permission to FBI Team automatically
+    """
+    dbx_sql = DBX_sql()
+    dbx_sql.grant_permissions_fbi(tbl_name)
 
-    try:
-        session = boto3.session.Session()
-        client = session.client(
-            service_name='secretsmanager',
-            region_name='us-east-1',
-            aws_access_key_id= key_id,
-            aws_secret_access_key= access_key,
-            # aws_session_token= os.getenv("AWS_SESSION_TOKEN")
-        )
+# TODO: drop table; might be easier to do it on UI, need to think about whether to delete the data file on s3 as well
 
-        response = client.get_secret_value(
-            SecretId=secret_name,
-        )
+def load_via_sql_dbx(load_df, tbl_name, local_path='./data', test_mode = True, s3_file = None):
+    """
+        This is the function that load pd df into databricks finance_accounting catalog
+        the function will save the pandas df as a parquet file and upload to s3 bucket (custom-upload/finance),
+        (the parquet file will be saved in data folder, or you can custimize the location by providing local_path parameter)
+        then databricks will read this parquet file and save as a delta table in finance_accounting catalog.
 
-        secrets = json.loads(response['SecretString'])
-        return secrets
-    except Exception as e:
-        print("Please double check settings.py file or environment vars!")
-        print(str(e))
-        return None
+        Parameters
+        ----------
+        load_df : pandas df to load
+        tbl_name : name you want the table to create
+        local_path : the location you want to save the parquet file
+        test_mode : default True, the table will be created in finance_test database; recommand leave as default
+        s3_file: the file name in s3 bucket, default None and will be the same as tbl_name; recommand leave as default
 
-def _decode_string(value):
-    return base64.b64decode(value).decode() #.replace("\n", "")
+        NOTE: local_path will be the same location where your entry point is
+    """
+    dbx_sql = DBX_sql()
+    dbx_sql.create_or_replace_table(load_df=load_df, tbl_name=tbl_name, local_path='./data', test_mode = True, s3_file = None)
 
-def read_value(value, encrypted=False):
-    return _decode_string(value) if encrypted else value #.replace("\n", "")
+def list_dbx_tables(catalog_name=None):
+    """
+        return a pandas dataframe showing all tables in one catalog
+        default catalog is finance_accounting if catalog_name is None, can also put like main, check databricks ui if needed 
+    """
+    dbx_sql = DBX_sql()
+    pdf = dbx_sql.list_all_tables(catalog_name=catalog_name)
+    return pdf
 
-# dbx_sql = DBX_sql()
-# def load_via_sql_dbx(load_df, tbl_name, secret_name = 'fbi_snowflake_creds', if_exists='replace', creds=None, test_mode=None):
-#     dbx_sql.create_or_replace_table()
+def query_method_by_env_dbx(query, use_service_account = True):
+    """
+        use in ETL pipeline
+        Load data through Spark API if on databricks;
+        use sql endpoint if in local environment;
+        default using service account to query to simplfying settings.py file saving on local
+    """
+    # query_databricks() using sql endpoint to load data, the performance is bad compare with query_snowflake (as of 4/22/2022)
+    # to improve performance, will need to install additional driver which is not available on dbx;
+    # therefore, using spark api (there is downside, causing warning for datatime schema!!)
 
+    # print is for test purpose, will remove when pr or next version (for user better understanding the logic)
+    if os.environ.get('environment') == 'databricks':
+        logging.debug('query via spark api, converting to pandas dataframe')
+        return query_via_spark_dbx(query)
+    else:
+        if use_service_account == True: logging.debug('query via sql endpoint, using service account creds')
+        else: logging.debug('query via sql endpoint, using personal account')
+        return query_databricks(query, use_service_account)
+    
 def load_method_by_env(value, key, exist_val, istest):
     """
         currently load value into both dbx and snowflakes
         dbx load using Spark API
         snowflake load using python sql connecter
-        TODO: will consider loading into only one environment
+        TODO: will consider loading into only one environment, when snowflake is retired
     """
+    if istest == False or istest == "False" or istest == None:
+        testflag = False
+    else: testflag = True
     if os.environ.get('environment') == 'databricks':
-        load_via_spark_dbx(value, key, exist_val, istest)
-    load_via_sql_snowflake(value, key, exist_val, istest)
+        load_via_spark_dbx(value, key, exist_val, istest=testflag)
+    load_via_sql_snowflake(load_df=value, tbl_name=key, if_exists=exist_val, test_mode=testflag)
+    # load_via_sql_dbx(load_df=value, tbl_name=key, test_mode=istest)
 
 def load_via_spark_dbx(value, key, exist_val, istest):
     """
@@ -722,8 +613,10 @@ def load_via_spark_dbx(value, key, exist_val, istest):
     sc2 = SparkContext.getOrCreate()
     spark_fbi = SparkSession(sc2)
     # spark_fbi.conf.set('spark.default.parallelism', '32')
+
     catalog = 'finance_accounting'
-    database = 'finance_test' if istest else 'finance_prod'
+    # in some case, istest is pass as string, changing in load_method_by_env() as well
+    database = 'finance_test' if istest==True else 'finance_prod'
     full_table_name = f'{catalog}.{database}.{key}'
     try:
         df = spark_fbi.table(full_table_name)
@@ -731,10 +624,11 @@ def load_via_spark_dbx(value, key, exist_val, istest):
     except AnalysisException as e:
         # print(str(e))
         databricks_table_exist_flag = False
-    pdf = value.astype(object).where(pd.notnull(value), None)
-    spark_schema = get_spark_schema(pdf)
+    # converting np.nan (which not accepted by spark) values to None
+    value = value.where(pd.notnull(value), None)
+    spark_schema = get_spark_schema(value)
     try:
-        df = spark_fbi.createDataFrame(pdf, spark_schema)
+        df = spark_fbi.createDataFrame(value, spark_schema)
         # if key == 'transaction_line_data': df = df.repartition(32)
         if databricks_table_exist_flag:
             if exist_val == 'replace': write_mode = 'overwrite'
@@ -744,26 +638,49 @@ def load_via_spark_dbx(value, key, exist_val, istest):
             df.write.format('delta') \
                 .option('path', f's3://di-databricks-production-finance/{database}/{key}') \
                 .saveAsTable(full_table_name)
-        print(f'{full_table_name} has been updated in DATABRICKS')
+            spark_fbi.sql(f'GRANT ALL PRIVILEGES ON TABLE {catalog}.{database}.{key} TO `FBI Team`')
+        logging.info(f'{full_table_name} has been updated in DATABRICKS')
     except Exception as e:
-        print(str(e))
+        logging.exception('Exception Occurred')
+
+def query_via_spark_dbx(query):
+    """
+        query through spark API, only for Spark env
+        ONLY for DBX, loading table via spark api
+    """
+    from pyspark.context import SparkContext
+    from pyspark.sql.session import SparkSession
+    from pyspark.sql.utils import AnalysisException
+    
+    sc2 = SparkContext.getOrCreate()
+    spark_fbi = SparkSession(sc2)
+    # TODO: this is causing error, because of some unsupported schema (datatime64?)
+    # spark_fbi.conf.set("spark.sql.execution.arrow.enabled","true")
+    df = spark_fbi.sql(query)
+    pdf = df.toPandas()
+    return pdf
 
 def get_spark_schema(pdf):
     """
         generate spark schema based on pandas dtypes
-        currently still require some manually fixing on certain column
+        currently still require some manually fixing on certain column; especially for Boolean type
     """
 
-    from pyspark.sql.types import StringType, IntegerType, DoubleType, DateType, BooleanType, StructField, StructType
+    from pyspark.sql.types import StringType, IntegerType, DoubleType, DateType, TimestampType, BooleanType, StructField, StructType
 
     struct_list = []
+    datetime64_cols = ['FinalClosingDate']
+
     for col, typ in zip(list(pdf.columns), list(pdf.infer_objects().dtypes)):
         
         if typ == 'object': spark_typ = StringType()
         elif typ == 'int64': spark_typ = IntegerType()
+        elif typ == 'datetime64[ns]' or typ == 'datetime64': spark_typ = TimestampType()
         else: spark_typ = DoubleType()
-            
-        if col.lower().endswith('date') or col.lower().endswith('period'): spark_typ = DateType()
+        
+
+        if (col.lower().endswith('date') or col.lower().endswith('period')) and (col.lower() not in [x.lower() for x in datetime64_cols]): 
+            spark_typ = DateType()
         if col == 'TRANSACTION_ID': spark_typ = IntegerType()
         if col == 'CLEAN_ADDRESS_PARSE': spark_typ = BooleanType()
         if col == 'same_market' or col == 'journal_entry_flag': spark_typ = BooleanType()
